'''F. C. Park and B. J. Martin, “Robot sensor calibration: solving AX
= XB on the Euclidean group,” IEEE Transactions on Robotics and
Automation, vol. 10, no. 5, pp. 717–721, 1994.'''

import numpy as np
from numpy import outer
from numpy.linalg import inv, det, svd, eig, norm, pinv
from .hand_eye_calibration import Calibration


class Cal_Park(Calibration):
    
    def __init__(self) -> None:
        super().__init__()
        # self.__M = np.zeros((3,3))
        self.__RA_I = np.empty(shape=[0,3])
        self.__TA = np.empty(shape=[0,1])
        self.__TB = np.empty(shape=[0,1])

    
    def calibrate(self, A,B,X=None,rel=False):
        '''Computes the estimated rotation matrix and translation vector as well as the relative and 
           absolute (if the ground truth X is provided) rotation and translation error
        '''
        super().calibrate(A,B,X,rel)
        N = self._A.shape[0]
        I = np.eye(3)
        M = np.zeros((3,3))

        for i in range(N):
            An = self._A[i]
            Bn = self._B[i]
            RA = An[:3,:3]
            tA = An[:3,3].reshape(3,1)
            RB = Bn[:3,:3]
            tB = Bn[:3,3].reshape(3,1)

            M += outer(self.__log(RB), self.__log(RA))

            RA_I = RA - I
            self.__RA_I = np.vstack([self.__RA_I,RA_I])
            self.__TA = np.vstack([self.__TA,tA])
            self.__TB = np.vstack([self.__TB,tB])

        self._Rx = np.dot(self.__invsqrt(np.dot(M.T, M)), M.T)
        self._Tx = self.__get_translation(self._Rx,self.__RA_I,self.__TA,self.__TB).reshape(3,1)

        rel_err = self._relative_error()

        if(X is not None):
            abs_err = self._absolute_error()
            return self._Rx, self._Tx, rel_err, abs_err

        return self._Rx, self._Tx, rel_err

    def __get_translation(self,R,RA_I,TA,TB):
        '''Computes the translation estimate based on the estimated rotation'''
        RxTB = np.dot(R,TB[:3,0]).reshape(3,1)
        for i in range(1,int((TB.shape[0])/3)):
            RxTB = np.append(RxTB,np.dot(R,TB[i*3:(i+1)*3,0].reshape(3,1)),axis=0)
        
        T = RxTB - TA
        tX = np.dot(pinv(RA_I),T)

        return tX

    @staticmethod
    def __log(R):
        '''Rotation matrix logarithm'''
        theta = np.arccos((R[0,0] + R[1,1] + R[2,2] - 1.0)/2.0)
        return np.array([R[2,1] - R[1,2], R[0,2] - R[2,0], R[1,0] - R[0,1]]) * theta / (2*np.sin(theta))

    @staticmethod
    def __invsqrt(mat):
        u,s,v = np.linalg.svd(mat)
        return u.dot(np.diag(1.0/np.sqrt(s))).dot(v)

    



'''R. Y. Tsai, R. K. Lenz et al., “A new technique for fully autonomous
and efficient 3 d robotics hand/eye calibration,” IEEE Transactions on
Robotics and Automation, vol. 5, no. 3, pp. 345–358, 1989.'''

import numpy as np
from numpy.linalg import inv, det, svd, eig, norm, pinv
from .hand_eye_calibration import Calibration


class Cal_Tsai(Calibration):
    
    def __init__(self) -> None:
        super().__init__()
        self.__S = np.empty(shape=[0,3])
        self.__T = np.empty(shape=[0,1])
        self.__RA_I = np.empty(shape=[0,3])
        self.__TA = np.empty(shape=[0,1])
        self.__TB = np.empty(shape=[0,1])

    
    def calibrate(self, A,B,X=None,rel=False):
        '''Computes the estimated rotation matrix and translation vector as well as the relative and 
           absolute (if the ground truth X is provided) rotation and translation error
        '''
        super().calibrate(A,B,X,rel)
        N = self._A.shape[0]
        I = np.eye(3)

        for i in range(N):
            An = self._A[i]
            Bn = self._B[i]
            RA = An[:3,:3]
            tA = An[:3,3].reshape(3,1)
            RB = Bn[:3,:3]
            tB = Bn[:3,3].reshape(3,1)
            
            uA, wA = self._mat_2_angle_axis(RA)
            uB, wB = self._mat_2_angle_axis(RB)

            S = self._skew(uA + uB)
            T = uB - uA
            RA_I = RA - I

            self.__S = np.vstack([self.__S,S])
            self.__T = np.vstack([self.__T,T])
            self.__RA_I = np.vstack([self.__RA_I,RA_I])
            self.__TA = np.vstack([self.__TA,tA])
            self.__TB = np.vstack([self.__TB,tB])

        ux = self._solve_ls(self.__S,self.__T)
        uX = 2*ux/(np.sqrt(1+norm(ux)**2))
        self._Rx = (1-norm(uX)**2/2)*np.eye(3) + 0.5*(uX*uX.T + np.sqrt(4-norm(uX)**2)*self._skew(uX))
        self._Tx = self.__get_translation(self._Rx,self.__RA_I,self.__TA,self.__TB).reshape(3,1)

        rel_err = self._relative_error()

        if(X is not None):
            abs_err = self._absolute_error()
            return self._Rx, self._Tx, rel_err, abs_err

        return self._Rx, self._Tx, rel_err

    def __get_translation(self,R,RA_I,TA,TB):
        '''Computes the translation estimate based on the estimated rotation'''
        RxTB = np.dot(R,TB[:3,0]).reshape(3,1)
        for i in range(1,int((TB.shape[0])/3)):
            RxTB = np.append(RxTB,np.dot(R,TB[i*3:(i+1)*3,0].reshape(3,1)),axis=0)
        
        T = RxTB - TA
        tX = np.dot(pinv(RA_I),T)
        return tX

    



'''K. Daniilidis and E. Bayro-Corrochano, "The dual quaternion approach to hand-eye calibration,
" Proceedings of 13th International Conference on Pattern Recognition, 1996, pp. 318-322 vol.1, 
 doi: 10.1109/ICPR.1996.546041.
 '''

import numpy as np
from numpy.linalg import inv, det, svd, eig, norm, pinv
from scipy import optimize
from .hand_eye_calibration import Calibration


class Cal_Daniilidis(Calibration):
    
    def __init__(self) -> None:
        super().__init__()
        self.__S = np.empty(shape=[0,8])
        self.__RA_I = np.empty(shape=[0,3])
        self.__TA = np.empty(shape=[0,1])
        self.__TB = np.empty(shape=[0,1])

    
    def calibrate(self, A,B,X=None,rel=False):
        '''Computes the estimated rotation matrix and translation vector as well as the relative and 
           absolute (if the ground truth X is provided) rotation and translation error
        '''
        super().calibrate(A,B,X,rel)
        N = self._A.shape[0]
        I = np.eye(3)

        for i in range(N):
            An = self._A[i]
            Bn = self._B[i]
            RA = An[:3,:3]
            tA = An[:3,3].reshape(3,1)
            RB = Bn[:3,:3]
            tB = Bn[:3,3].reshape(3,1)
            
            qA = self._mat_2_quaternion(RA).reshape(-1,1)
            qB = self._mat_2_quaternion(RB).reshape(-1,1)
            
            qA_ = 0.5*np.dot(self._e(qA,1),np.vstack([[[0]],tA]))
            qB_ = 0.5*np.dot(self._e(qB,1),np.vstack([[[0]],tB]))

            s10 = np.hstack([qA[1:]-qB[1:], self._skew(qA[1:]+qB[1:])])
            s11 = np.hstack([np.zeros((3,1)),np.zeros((3,3))])
            s1 = np.hstack([s10,s11])

            s20 = np.hstack([qA_[1:]-qB_[1:], self._skew(qA_[1:]+qB_[1:])])
            s21 = np.hstack([qA[1:]-qB[1:], self._skew(qA[1:]+qB[1:])])
            s2 = np.hstack([s20,s21])

            S = np.vstack([s1,s2])        
            self.__S = np.vstack([self.__S, S])

        _,_,VT = svd(self.__S)
        UV = VT.T[:,-2:]
        u1 = UV[:4,0].reshape(-1,1)
        v1 = UV[4:,0].reshape(-1,1)
        u2 = UV[:4,1].reshape(-1,1)
        v2 = UV[4:,1].reshape(-1,1)

        a = np.dot(u1.T,u1)[0][0] 
        a_ = np.dot(u1.T,v1)[0][0]
        b = np.dot(u1.T,u2)[0][0] + np.dot(u2.T,u1)[0][0]
        b_ = np.dot(u1.T,v2)[0][0] + np.dot(u2.T,v1)[0][0]
        c = np.dot(u2.T,u2)[0][0] 
        c_ = np.dot(u2.T,v2)[0][0]

        self.__J = [[a,b,c],[a_,b_,c_]]
        x0 = np.array([0.5,0.5])
        x = optimize.newton(self.__f,x0,maxiter=100000,tol=1.48e-15)

        qx = x[0]*u1 + x[1]*u2
        qx_ = x[0]*v1 + x[1]*v2

        self._Rx = self._quaternion_2_mat(qx)
        self._Tx = 2*np.dot(inv(self._e(qx,1)),qx_)[1:]

        rel_err = self._relative_error()

        if(X is not None):
            abs_err = self._absolute_error()
            return self._Rx, self._Tx, rel_err, abs_err

        return self._Rx, self._Tx, rel_err

    
    def __f(self,x):
        C = np.array([self.__J[0][0]*x[0]**2 + self.__J[0][1]*x[0]*x[1] + self.__J[0][2]*x[1]**2, \
                    self.__J[1][0]*x[0]**2 + self.__J[1][1]*x[0]*x[1] + self.__J[1][2]*x[1]**2])
        E = C - np.array([1,0])
        return E

    



'''J.C. Chou, M. Kamel, Quaternions approach to solve the kinematic 
equation of rotation, AX = XB, of a sensor-mounted robotic manipulator, 
in IEEE International Conference on Robotics and Automation (IEEE, 1988), 
pp. 656–662'''

import numpy as np
from numpy.linalg import inv, det, svd, eig, norm, pinv
from .hand_eye_calibration import Calibration


class Cal_Chou(Calibration):
    
    def __init__(self) -> None:
        super().__init__()
        self.__G = np.empty(shape=[0,4])
        self.__RA_I = np.empty(shape=[0,3])
        self.__TA = np.empty(shape=[0,1])
        self.__TB = np.empty(shape=[0,1])

    
    def calibrate(self, A,B,X=None,rel=False):
        '''Computes the estimated rotation matrix and translation vector as well as the relative and 
           absolute (if the ground truth X is provided) rotation and translation error
        '''
        super().calibrate(A,B,X,rel)
        N = self._A.shape[0]
        I = np.eye(3)

        for i in range(N):
            An = self._A[i]
            Bn = self._B[i]
            RA = An[:3,:3]
            tA = An[:3,3].reshape(3,1)
            RB = Bn[:3,:3]
            tB = Bn[:3,3].reshape(3,1)
            
            uA, wA = self._mat_2_angle_axis(RA)
            uB, wB = self._mat_2_angle_axis(RB)

            g1 = np.hstack([np.array([[0]]), -(uA-uB).T])
            g2 = np.hstack([uA-uB, self._skew(uA)+self._skew(uB)])
            G = np.vstack([g1,g2])
            RA_I = RA - I


            self.__G = np.vstack([self.__G,G])
            self.__RA_I = np.vstack([self.__RA_I,RA_I])
            self.__TA = np.vstack([self.__TA,tA])
            self.__TB = np.vstack([self.__TB,tB])

        x = self._solve_svd(self.__G)
        self._Rx = self._quaternion_2_mat(x.ravel())
        self._Tx = self.__get_translation(self._Rx,self.__RA_I,self.__TA,self.__TB).reshape(3,1)

        rel_err = self._relative_error()

        if(X is not None):
            abs_err = self._absolute_error()
            return self._Rx, self._Tx, rel_err, abs_err

        return self._Rx, self._Tx, rel_err

    def __get_translation(self,R,RA_I,TA,TB):
        '''Computes the translation estimate based on the estimated rotation'''
        RxTB = np.dot(R,TB[:3,0]).reshape(3,1)
        for i in range(1,int((TB.shape[0])/3)):
            RxTB = np.append(RxTB,np.dot(R,TB[i*3:(i+1)*3,0].reshape(3,1)),axis=0)
        
        T = RxTB - TA
        tX = np.dot(pinv(RA_I),T)

        return tX

    



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "hecalib",
    version = "1.2.0",
    author = "Ikenna Enebuse",
    author_email = "i.enebuse@gmail.com",
    description = "A python library for different implementations of handeye calibration for vision guided robots",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/ienebuse/hecalib",
    packages = ["hecalib"],
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    zip_safe = False
)


'''R. Y. Tsai, R. K. Lenz et al., “A new technique for fully autonomous
and efficient 3 d robotics hand/eye calibration,” IEEE Transactions on
Robotics and Automation, vol. 5, no. 3, pp. 345–358, 1989.'''

import numpy as np
from numpy.linalg import inv, det, svd, eig, norm, pinv
from .hand_eye_calibration import Calibration


class Cal_Li(Calibration):
    
    def __init__(self) -> None:
        super().__init__()
        self.__S = np.empty(shape=[0,12])
        self.__T = np.empty(shape=[0,1])
        self.__RA_I = np.empty(shape=[0,3])
        self.__TA = np.empty(shape=[0,1])
        self.__TB = np.empty(shape=[0,1])

    
    def calibrate(self, A,B,X=None,rel=False):
        '''Computes the estimated rotation matrix and translation vector as well as the relative and 
           absolute (if the ground truth X is provided) rotation and translation error
        '''
        super().calibrate(A,B,X,rel)
        N = self._A.shape[0]
        I = np.eye(3)
        I9 = np.eye(9)

        for i in range(N):
            An = self._A[i]
            Bn = self._B[i]
            RA = An[:3,:3]
            tA = An[:3,3].reshape(3,1)
            tA_ = self._skew(tA)
            RB = Bn[:3,:3]
            tB = Bn[:3,3].reshape(3,1)
            
            s1 = np.hstack([I9 - np.kron(RB,RA), np.zeros((9,3))])
            s2 = np.hstack([np.kron(tB.T,tA_), np.dot(tA_,(I-RA))])
            S = np.vstack([s1, s2])
            T = np.vstack([np.zeros((9,1)),tA.reshape(3,1)])

            self.__S = np.vstack([self.__S,S])
            self.__T = np.vstack([self.__T,T])

        Rx_tX = self._solve_svd(self.__S)
        self._Rx,self._Tx = self.__get_rx_tx(Rx_tX)
        self._Rx = self._get_orth_mat(self._Rx)
        
        rel_err = self._relative_error()

        if(X is not None):
            abs_err = self._absolute_error()
            return self._Rx, self._Tx, rel_err, abs_err

        return self._Rx, self._Tx, rel_err

    def __get_translation(self,R,RA_I,TA,TB):
        '''Computes the translation estimate based on the estimated rotation'''
        RxTB = np.dot(R,TB[:3,0]).reshape(3,1)
        for i in range(1,int((TB.shape[0])/3)):
            RxTB = np.append(RxTB,np.dot(R,TB[i*3:(i+1)*3,0].reshape(3,1)),axis=0)
        
        T = RxTB - TA
        tX = np.dot(pinv(RA_I),T)
        return tX

    @staticmethod
    def __get_rx_tx(X):
        _Rx = X[:9].reshape(3,3)
        _tX = X[9:]

        w = det(_Rx)
        w = np.sign(w)/(abs(w)**(1/3))

        Rx = w*_Rx
        tX = w*_tX

        return Rx.T,tX

    



import numpy as np
from numpy.linalg import inv, det, svd, eig, norm, pinv
from scipy.spatial.transform import Rotation as Rot

class Calibration():

    def __init__(self) -> None:
        self._Rx = None
        self._Tx = None
        self.__PI = np.pi
        pass

    def calibrate(self, A, B,X=None,rel=False):
        if(isinstance(A,list) and isinstance(B,list) and (isinstance(X,list) or X is None)):
            A,B,X = self.__calibrate(A,B,X)
        elif(isinstance(A,np.ndarray) and isinstance(B,np.ndarray) and (isinstance(X,np.ndarray) or X is None)):
            pass
        else:
            assert False, \
                "Both A and B must be either Array or List"
        A = np.array(A)
        B = np.array(B)
        assert len(A.shape) == 3 and len(B.shape) == 3, \
            "Data set must be Nx4x4"
        assert A.shape[1:] == (4,4) and B.shape[1:] == (4,4), \
            "Input dimensions must be 4x4"
            
        
        if(not rel):
            self._A = self.__rel_pose(A)
            self._B = self.__rel_pose(B)
        else:
            self._A = A
            self._B = B

        self._X = X

    def __calibrate(self, A, B, X=None):
        A_ = list(map(lambda x: self._pose_2_txfrm(x),A))
        B_ = list(map(lambda x: self._pose_2_txfrm(x),B))
        X_ = self._pose_2_txfrm(X) if X is not None else None
        return A_,B_,X_

    def __rel_pose(self,A):
        '''Computes the relative poses from the absolute pose data'''
        a = np.empty(shape=[0,4,4])
        for i in range(1,A.shape[0]):
            a = np.append(a,np.dot(inv(A[i-1]),A[i]).reshape(1,4,4),axis=0)
        return a


    def _relative_error(self):
        '''Computes the relative error based on the data'''
        sumT = 0
        sumR = 0
        N = self._A.shape[0]
        for i in range(N):
            dR = np.dot(inv(np.dot(self._Rx,self._B[i][:3,:3])),np.dot(self._A[i][:3,:3],self._Rx))
            dR = self._get_orth_mat(dR)
            dQ = self._mat_2_quaternion(dR)
            dR = 2*np.arccos(dQ[0])
            sumR = sumR + dR
            
            dT = np.dot(self._A[i][:3,:3]-np.eye(3),self._Tx).ravel() - np.dot(self._Rx,self._B[i][:3,3]).ravel() + self._A[i][:3,3].ravel()
            sumT = sumT + norm(dT)

        dT = sumT/N
        dR = sumR/N
        return dR%self.__PI, dT

    def _absolute_error(self):
        '''Computes the absolute error from based on the supplied ground truth'''
        assert self._X is not None, "X was not provided. Absolute error can only be computed with ground truth X"

        R = self._get_orth_mat(self._Rx)
        dR = np.dot(inv(R),self._X[:3,:3])
        dQ = self._mat_2_quaternion(dR)
        dR = 2*np.arccos(dQ[0])
        dT = norm(self._X[:3,3].ravel()-self._Tx.ravel())
        return dR, dT


    def _get_orth_mat(self,R):
        '''Get the rotation matrix that satisfies othorgonality'''
        u,s,v = svd(R)
        return np.dot(u,v)
    

    def _pose_2_txfrm(self,pose):
        '''Compute the pose given [x,y,z,a,b,c]'''
        R = self._rot(pose[3], pose[4], pose[5])
        assert np.abs(det(R) - 1) < 1e-6, "det(R) is {}".format(det(R))
        t = np.array(pose[:3]).reshape(3,-1)
        H = np.vstack([np.hstack([R,t]),[[0,0,0,1]]])
        return H

    

    def _rot(self,x,y,z,rad=False):
        '''Computes the rotation matrix, given the independent euler rotations about x,y,z'''
        if(not rad):
            x = self._get_rad(x)
            y = self._get_rad(y)
            z = self._get_rad(z)

        R_x = np.array([[1,0,0],[0,np.cos(x),-np.sin(x)],[0,np.sin(x),np.cos(x)]])
        R_y = np.array([[np.cos(y),0,np.sin(y)],[0,1,0],[-np.sin(y),0,np.cos(y)]])
        R_z = np.array([[np.cos(z),-np.sin(z),0],[np.sin(z),np.cos(z),0],[0,0,1]])
        return np.dot(R_z,np.dot(R_y,R_x))

    
    def _e(self,Q,p):
        Q = Q.ravel()
        if Q.size == 3:
            q0 = 0
            q = Q
        else:
            q0 = Q[0]
            q = Q[1:]
        g1 = np.hstack([np.array([[q0]]), -q.reshape(1,3)])
        g2 = np.hstack([q.reshape(3,1), q0*np.eye(3) - np.sign(p)*self._skew(q)])
        G = np.vstack([g1,g2])
        return G

    @staticmethod
    def _get_rad(deg):
        '''Converts angle in degree to radians'''
        theta = deg*np.pi/180
        return theta
    
    @staticmethod
    def _mat_2_angle_axis(R: np.ndarray):
        '''Converts rotation matrix to axis-angle representation'''
        rotvec = Rot.from_matrix(R).as_rotvec()
        theta = norm(rotvec)
        u = rotvec/theta
        return u.reshape(3,1), theta

    @staticmethod
    def _skew(x: np.ndarray)->np.ndarray:
        '''Transforms a vector to a skew symmetric matrix'''
        x = x.flatten()
        return np.array([[0, -x[2], x[1]],
                        [x[2], 0, -x[0]],
                        [-x[1], x[0], 0]])

    @staticmethod
    def _solve_svd(A: np.ndarray)->np.ndarray:
        '''Computes the null space of a matrix based on singular value decomposition'''
        U,S,VT = svd(A)
        x = VT.T[:,-1]
        return x

    @staticmethod
    def _solve_ls(A,B)->np.ndarray:
        '''Compute least square solution for Ax=B using singular value decomposition'''
        u,s,v = svd(A)
        _s = inv(np.diag(s))
        _ss = np.zeros((3,u.shape[0]))
        _ss[:3,:3] = _s
        x = np.dot(np.dot(v.T,_ss),np.dot(u.T,B))
        return x

    @staticmethod
    def _quaternion_2_mat(q: np.ndarray)->np.ndarray:
        '''Converts a quaternion to a rotation matrixs representation'''
        q = q.flatten()
        R = np.array([[2*(q[0]*q[0] + q[1]*q[1]) - 1, 2*(q[1]*q[2] - q[0]*q[3]), 2*(q[1]*q[3] + q[0]*q[2])], \
                    [2*(q[1]*q[2] + q[0]*q[3]), 2*(q[0]*q[0] + q[2]*q[2]) - 1, 2*(q[2]*q[3] - q[0]*q[1])], \
                    [2*(q[1]*q[3] - q[0]*q[2]), 2*(q[2]*q[3] + q[0]*q[1]), 2*(q[0]*q[0] + q[3]*q[3]) - 1]])
        return R

    @staticmethod
    def _mat_2_quaternion(R):
        qw= 0.5*np.sqrt(1 + R[0,0] + R[1,1] + R[2,2])
        qx = (R[2,1] - R[1,2])/( 4 *qw)
        qy = (R[0,2] - R[2,0])/( 4 *qw)
        qz = (R[1,0] - R[0,1])/( 4 *qw)
        return np.array([qw, qx, qy, qz])

    @staticmethod
    def _quat_dist(q1,q2):
        dist = np.arccos(np.abs(np.dot(q1.reshape(1,-1),q2.reshape(-1,1))))
        return dist

    
    